from unittest import TestCase


class TestKaggleKernel(TestCase):
    def test_install_dependency(self):
        self.fail()

    def test_set_logger(self):
        self.fail()

    def test_set_random_seed(self):
        self.fail()

    def test_save_model(self):
        self.fail()

    def test_train_model(self):
        self.fail()

    def test_set_model(self):
        self.fail()

    def test_prepare_train_dev_data(self):
        self.fail()

    def test_prepare_test_data(self):
        self.fail()

    def test_run(self):
        self.fail()

    def test_continue_run(self):
        self.fail()

    def test_submit(self):
        self.fail()
